/*
 *  Copyright (c) 1993-2002 Easysoft Ltd. All rights reserved.
 *  
 *  xmlodbc_client
 *  ==============
 *
 *  A simple client for the Easysoft XML-ODBC Server.
 *
 *  Command line:
 *
 *  xmlodbc_client -h server_host
 *                 [-p server_port]
 *                 [-s "SQL query" -c "connection string"] | [ -f request file ]
 *
 *  e.g. xmlodbc_client -h myserver -p 8895
 *                      -s "select * from table"
 *                      -c "DSN=test;UID=dbname;PWD=dbpassword"
 *
 *  Output is to stdout and in XML.
 *
 *  $Id: xmlodbc_client.c,v 1.5 2002/07/12 14:20:31 martin Exp $
 *
 *  $Log: xmlodbc_client.c,v $
 *  Revision 1.5  2002/07/12 14:20:31  martin
 *  Ignore SIGPIPE.
 *
 *  Revision 1.4  2002/07/08 07:39:24  martin
 *  Handle change to server which supports multiple requests.
 *
 *  Revision 1.3  2002/05/07 10:40:42  martin
 *  Handle CDATA.
 *
 *  Revision 1.2  2002/05/07 08:30:53  martin
 *  Various command line parsing fixes.
 *  Add support for -f request_file
 *
 *  Revision 1.1  2002/04/12 09:43:34  martin
 *  Initial revision
 *
 *
 */
#ifdef WIN32
# include <winsock.h>
# include <windows.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

#ifndef WIN32
# include <netdb.h>
# include <sys/socket.h>
# include <netinet/in.h>
extern int h_errno;
#endif

#include <sys/types.h>

#include "config.h"

#ifdef HAS_SIGNALS
# include <signal.h>
#endif

static void usage(void);
static int request(char *host, char *port, char *connstr, char *sql,
                   char *file);
static int start_winsock(void);
static void stop_winsock(void);
static int read_request_from_file(char *file, char **buf, int *buflen);
static char *escape_string(char *instr);
#ifdef HAS_SIGNALS
static void ignore_sigpipe(void);
#endif

#ifdef WIN32
static struct WSAData           wsadata;        /* WinSock */
#endif


/************************************************************************/
/*                                                                      */
/*  main                                                                */
/*  ====                                                                */
/*                                                                      */
/*  Parse the command line and despatch the request.                    */
/*                                                                      */
/************************************************************************/
main(
    int argc,
    char *argv[])
{
    int                 arg;                    /* arg in argv */
    char                *sql=NULL;              /* sql query */
    char                *port=NULL;             /* port string */
    char                *host=NULL;             /* name of server host */
    char                *connstr=NULL;          /* ODBC connection string */
    char                *file=NULL;             /* request in file */
    int                 verbose=0;              /* verbose output */
    char                *escsql;                /* escaped version of sql */
    char                *escconnstr;            /* escaped version of connstr */
    
    arg = 1;
    while(arg < argc) {
        char    schar;
        
        if (argv[arg][0] == '-') {
            if (strlen(argv[arg]) <= 1) {
                fprintf(stderr, "Missing option\n");
                usage();
                exit(1);
            }
            schar = argv[arg][1];
            if (schar != 'v') {
                arg++;
                if (arg >= argc) {
                    fprintf(stderr, "Missing argument for %c\n", schar);
                    usage();
                    exit(3);
                }
            }
            
            switch(schar)
            {
              case 'v':
              {
                  verbose = 1;
                  break;
              }
              case 'h':
              {
                  host = argv[arg];
                  break;
              }
              case 'p':
              {
                  port = argv[arg];
                  break;
              }
              case 'c':
              {
                  connstr = argv[arg];
                  break;
              }
              case 's':
              {
                  sql = argv[arg];
                  break;
              }
              case 'f':
              {
                  file = argv[arg];
                  break;
              }
            }
            arg++;
        }
        else {
            fprintf(stderr, "Parse error at \'%s\'\n", argv[arg]);
            usage();
            exit(2);
        }
    }
    if (!host ||
        (!file && (!sql || !connstr)) ||
        (file && (sql || connstr)))
    {
        fprintf(stderr, "Invalid option combination\n");
        usage();
        exit(4);
    }
    if (!port) port = DEFAULT_PORT_STRING;

    if (!file) {
        escsql = escape_string(sql);
        escconnstr = escape_string(connstr);

        if (!escsql || !escconnstr) {
            fprintf(stderr,
                    "Failed to allocate memory for escaped "
                    "sql or connection string\n");
            exit(4);
        }
    }
    
    if (verbose) {
        printf("Host: %s\n", host);
        if (file) {
            printf("File: %s\n", file);
        }
        else {
            printf("Connection string: %s\n", escconnstr);
            printf("SQL: %s\n", escsql);
        }
        printf("Port: %s\n", port);
    }
    
    request(host, port, escconnstr, escsql, file);

    if (!file) {
        free(escsql);
        free(escconnstr);
    }
    
}


/************************************************************************/
/*                                                                      */
/*  request                                                             */
/*  =======                                                             */
/*                                                                      */
/************************************************************************/
static int request(
    char *host,
    char *port,
    char *connstr,
    char *sql,
    char *file)
{
    int                 nport;                  /* numeric port */
    struct hostent      *he;                    /* host entry */
    unsigned long       inetaddr;               /* inet address */
    int                 sock;                   /* socket for comms */
    struct sockaddr_in  sa;                     /* address to connect to */
    int                 sts;                    /* fn return status */
    char                *ibuf=NULL;             /* input buffer for request */
    int                 ibuflen=0;              /* ... and its length */
    char                *p;                     /* ptr into ibuf */
    int                 bytes_sent;             /* bytes sent in request */
    char                obuf[1025];             /* output buffer */
    
    if (file) {
        sts = read_request_from_file(file, &ibuf, &ibuflen);
        if (sts < 0) return sts;
    }
    
    if (start_winsock() < 0) return -1;

#ifdef HAS_SIGNALS
    ignore_sigpipe();
#endif
    
    nport = atoi(port);
    if (!(he = gethostbyname(host))) {
        fprintf(stderr, "Failed to lookup %s - error %d\n",
                host, WLASTERROR);
        return -1;
    }
    
    memcpy(&inetaddr, he->h_addr_list[0], sizeof(inetaddr));

    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock == -1) {
        fprintf(stderr, "Failed to create socket - %s\n",
                strerror(errno));
        return -2;
    }
    
    memset(&sa, 0, sizeof(sa));
    sa.sin_family = AF_INET;
    sa.sin_port = htons(nport);
    memcpy(&sa.sin_addr, &inetaddr, sizeof(inetaddr));
    
    sts = connect(sock, (struct sockaddr *)&sa, sizeof(sa));
    if (sts == -1) {
        fprintf(stderr, "Failed to connect - %s\n",
                strerror(errno));
        return -3;
    }

    if (!file) {
        ibuflen= strlen(connstr) + strlen(sql) + 1024;
        if (!(ibuf = malloc(ibuflen))) {
            fprintf(stderr, "Failed to allocate %d bytes for input buffer\n",
                    ibuflen);
            return -4;
        }
        p = ibuf;
        p += sprintf(p, "<?xml version=\"1.0\"?>\r\n");
        p += sprintf(p, "<request>\r\n");
        p += sprintf(p, "<connectionstring>%s</connectionstring>\r\n", connstr);
        p += sprintf(p, "<sql>%s</sql>\r\n", sql);
        p += sprintf(p, "</request>\r\n");
        ibuflen = strlen(ibuf);
    }
    
    p = ibuf;
    bytes_sent = 0;
    while(bytes_sent < ibuflen) {
        sts = send(sock, ibuf, strlen(ibuf), 0);
        if (sts == -1) {
            fprintf(stderr, "Failed to send request - %s\n",
                    strerror(errno));
            free(ibuf);
            return -5;
        }
        bytes_sent += sts;
    }
    shutdown(sock, 1);

    do {
        sts = recv(sock, obuf, sizeof(obuf) - 1, 0);
        if (sts > 0) {
            obuf[sts] = '\0';
            if (strstr(obuf, "</Result>")) {
                fwrite(obuf, sts, 1, stdout);
                break;
            }
        }
        
        fwrite(obuf, sts, 1, stdout);
    } while(sts > 0);
    shutdown(sock, 0);
    
    CLOSESOCKET(sock);
    free(ibuf);
    stop_winsock();
    
    return 0;
    
}


/************************************************************************/
/*                                                                      */
/*  usage                                                               */
/*  =====                                                               */
/*                                                                      */
/************************************************************************/
static void usage(void)
{
    printf("xmlodbc_client "
           "[-v] -h host  [-p port]\n"
           "\t[-s sql -c connection_string] | [-f file]\n");
    return;
}


/************************************************************************/
/*                                                                      */
/*  start_winsock                                                       */
/*  =============                                                       */
/*                                                                      */
/************************************************************************/
static int start_winsock(void)
{
#ifdef WIN32
    WORD wVersionRequested;
    int err;

    /*
     *  Request WinSock 2.0
     */
    wVersionRequested = MAKEWORD(2, 0);         /* 1.1 worked on new win95*/

    err = WSAStartup(wVersionRequested, &wsadata);
    if (err != 0)
    {
        fprintf(stderr, "WSAStartup fails requesting 2 0\n");
        return -1;
    }

    /*
     *  Confirm that the Windows Sockets DLL supports 2.0
     *  Note that if the DLL supports versions greater than 2.0 in addition
     *  to 2.0, it will still return 2.0 in wVersion since that is the version
     *  we requested.
     */
    if (LOBYTE(wsadata.wVersion) != 2 || HIBYTE(wsadata.wVersion) != 0)
    {
        fprintf(stderr, "Winsock mismatch %d %d\n",
                LOBYTE(wsadata.wVersion), HIBYTE(wsadata.wVersion));
    	WSACleanup();
    	return -1; 
    }
#endif /* WIN32 */
    
    return 0;
}

/************************************************************************/
/*                                                                      */
/*  stop_winsock                                                        */
/*  ============                                                        */
/*                                                                      */
/************************************************************************/
static void stop_winsock(void)
{
#ifdef WIN32
    WSACleanup();
#endif
}


/************************************************************************/
/*                                                                      */
/*  read_request_from_file                                              */
/*  ======================                                              */
/*                                                                      */
/************************************************************************/
static int read_request_from_file(
    char *file,
    char **buf,
    int *buflen
    )
{
    FILE                *fp=NULL;               /* request file */
    size_t              bytes_read;             /* bytes read from file */
    char                *p;                     /* ptr into ibuf */
    char                *ibuf = NULL;
    int                 ibuflen;
    
    fp = fopen(file, "rb");
    if (!fp) {
        fprintf(stderr, "Failed to open %s for read access - %s\n",
                strerror(errno));
        return -1;
    }
    if (!(ibuf = malloc(1024))) {
            fprintf(stderr, "Failed to allocate 1024 bytes\n");
            return -1;
    }
    ibuflen = 0;
    p = ibuf;
    do {
        bytes_read = fread(p, 1, 1024, fp);
        ibuflen += bytes_read;
        if (bytes_read == 1024) {
            if (!(ibuf = realloc(ibuf, ibuflen + 1024))) {
                fprintf(stderr,
                        "Failed to allocate %d bytes for input buffer\n",
                        ibuflen + 1024);
                return -1;
            }
        }
        p += bytes_read;
    } while(bytes_read);

    *buf = ibuf;
    *buflen = ibuflen;
    
    return 0;
    
}


/************************************************************************/
/*                                                                      */
/*  escape_string                                                       */
/*  =============                                                       */
/*                                                                      */
/*  Escape &, < and >.                                                  */
/*                                                                      */
/************************************************************************/
static char *escape_string(
    char *instr)
{
    char        *escstr;
    char        *src;
    char        *dst;

    /*
     *  If the string is CDATA leave it alone and return a duplicate.
     */
    if (strstr(instr, "<![CDATA[") == instr) return strdup(instr);
    
    /*
     *  & => &amp;
     *  > => &gt;
     *  < => &lt;
     */
    if (!(escstr = malloc(strlen(instr) * 5))) return NULL;

    src = instr;
    dst = escstr;
    while(*src) {
        if (*src == '<') {
            strcpy(dst, "&lt;");
            dst += 4;
        } else if (*src == '>') {
            strcpy(dst, "&gt;");
            dst += 4;
        } else if (*src == '&') {
            strcpy(dst, "&amp;");
            dst += 5;
        } else {
            *dst++ = *src;
        }
        src++;
    }
    *dst = '\0';
    return escstr;
}

#ifdef HAS_SIGNALS
/************************************************************************/
/*                                                                      */
/*  ignore_sigpipe                                                      */
/*  ==============                                                      */
/*                                                                      */
/*  Ignore SIGPIPE so we get EPIPE error instead.                       */
/*                                                                      */
/************************************************************************/
static void ignore_sigpipe(void)
{
    struct sigaction    act;
    sigset_t            sa_mask;

    sigemptyset(&sa_mask);
    act.sa_handler = SIG_IGN;
    act.sa_mask = sa_mask;
    act.sa_flags = SA_RESTART;

    sigaction(SIGPIPE, &act, NULL);
}

#endif /* HAS_SIGNALS */
