package ESXMLODBC::XMLREQUEST;
#
# Copyright (c) 1993-2002 Easysoft Ltd. All rights reserved.
#
# $Id: XMLREQUEST.pm,v 1.10 2002/07/12 15:17:59 martin Exp $
#
# This is the Perl package to allow a request to be sent to the Easysoft
# XML-ODBC Server and retrieve the results.
#
# You may use or alter this package as you see fit however easysoft
# would appreciate any feedback - comments or patches. These may be
# sent to support@easysoft.com but please mark these as
# "ESXMLODBC::XMLREQUEST" to ensure a response.
#
# If you set Parse then you need XML::Parser. XML::Parser needs expat
# at a minimum but it is best if you also have libwww. Here is the tree
# of dependencies:
#
# XML-Parser
#   expat
#   libwww
#     URI
#       MIME::Base64
#     HTML::Parser
#     MIME::Base64
#     Net::FTP  (you can find this in libnet)
# 
use strict;
require 5.004;

BEGIN {
    $ESXMLODBC::VERSION = 1.0;
}

=head1 NAME

ESXMLODBC::XMLREQUEST - XML to ODBC

=head1 SYNOPSIS

  use ESXMLODBC::XMLREQUEST;

  $h = ESXMLODBC::XMLREQUEST->new();

  $h->Connect();
  $h->Connect(\%attr);

  $h->Disconnect();

  $h->ExecuteFile($file);
  $h->Execute();
  $h->Execute($sql);
  $h->Execute($sql, \%attr);

  $h->First();
  $h->Next();
  $h->Last();

=cut

# POD continues at end

use strict;
use IO::Socket;

my $_self;

my %resultrowhash;		# current row
my @elem_stack;

#
# new
#
sub new {
    my $class = shift;
    my $self = {};
    bless ($self, $class);

    $self->{Port} = 8895;
    $self->{ConnectionString} = "";
    $self->{Sql} = undef;
    $self->{Parse} = 1;
    $self->{Truncate} = 0;
    $self->{_executed} = 0;
    $self->{_pos} = 0;
    $_self = $self;
    _init_state($self);
    return $self;
}

#
# Connect
#
sub Connect {
    my $self = shift;
    my $hashargs = shift;

    _init_state($self);
    if ($hashargs) {
	my $mkey;
	foreach $mkey (keys %$hashargs) {
	    $self->{$mkey} = $hashargs->{$mkey};
	}
    }
    if ($self->{_Connected}) {
	$self->{ErrorState} = "HY000";
	$self->{ErrorText} = "Already connected";
	return undef;
    }
    if (!exists($self->{Server})) {
	$self->{ErrorState} = "HY000";
	$self->{ErrorText} = "No server specified";
	return undef;
    }
    
    $self->{_socket} = 
	IO::Socket::INET->new(
			      PeerAddr => $self->{Server},
			      PeerPort => $self->{Port},
			      Proto => "tcp",
			      Timeout => 20,
			      Type => SOCK_STREAM);
    if (!$self->{_socket}) {
	$self->{ErrorState} = "HY000";
	$self->{ErrorText} = 
	    "Failed to connect to server $self->{Server} on port $self->{Port}";
	return undef;
    }
    $self->{_Connected} = 1;
}
    
#
# Disconnect
#
sub Disconnect {
    my $self = shift;
    
    _init_state($self);
    close($self->{_socket}) if ($self->{_Connected});
    $self->{_Connected} = undef;
}

#
# ExecuteFile
#
sub ExecuteFile {
    my $self = shift;
    my $file = shift;

    _init_execute($self);

    if (!$self->{_Connected}) {
	$self->{ErrorState} = "HY000";
	$self->{ErrorText} = "Not connected";
	return undef;
    }

    if (!open (FH, "<$file")) {
	$self->{ErrorState} = "HY000";
	$self->{ErrorText} = "Failed to open file - $!";
	return undef;
    }

    my $sock = $self->{_socket};

    print $sock <FH>;
    close (FH);

    my $res;
    while($res = <$sock>) {
	$self->{XML} .= $res;
	last if ($res =~ /<\/Result>/);
    }
    $self->{_executed} = 1;
    $self->{_pos} = 0;

    return 1 if (!$self->{Parse});

    return _parse($self);

    return 1;
}


#
# Execute
#
sub Execute {
    my $self = shift;
    my $arg = shift;
    my $sql=undef;

    _init_execute($self);

    if (!ref($arg)) {
	$sql = $arg;
	$arg = shift;
    }
    if ($arg) {
	my $mkey;
	foreach $mkey (keys %$arg) {
	    $self->{$mkey} = $arg->{$mkey};
	}
    }

    if (!$self->{_Connected}) {
	$self->{ErrorState} = "HY000";
	$self->{ErrorText} = "Not connected";
	return undef;
    }
    if (!$self->{ConnectionString}) {
	$self->{ErrorState} = "HY000";
	$self->{ErrorText} = "Missing ConnectionString";
	return undef;
    }
    if (!$self->{Sql} && !$sql) {
	$self->{ErrorState} = "HY000";
	$self->{ErrorText} = "Missing SQL";
	return undef;
    }
    #
    # Form the XML request to send to the server.
    #
    my $sxml = 	
	"<?xml version=\"1.0\"?>\r\n<Request>\r\n  <ConnectionString>" . 
	$self->{ConnectionString} . "  </ConnectionString>\r\n";
    # escape SQL
    if ($sql) {
	$sql = _escape($sql);
	$sxml .= "  <Sql>$sql</Sql>\r\n";
    } else {
	$sxml .= "  <Transaction>\r\n";
	my $n = 0;
	while (defined($self->{Sql}->[$n])) {
	    $sql = _escape($self->{Sql}->[$n]);
	    $sxml .= "    <Sql>" . $sql . "</Sql>\r\n";
	    $n++;
	}
	$sxml .= "  </Transaction>\r\n";
    }
    $sxml .= "  <Format>\r\n    <Long/>\r\n";
    if ($self->{Truncate}) {
	$sxml .= "    <TruncateTrailingBlanks/>\r\n";
    }
    $sxml .= "  </Format>\r\n";
    $sxml .= "</Request>\r\n";
    $self->{Request} = $sxml;

    $self->{XML} = "";

    my $sock = $self->{_socket};

    print $sock $sxml;
    
    my $res;
    while($res = <$sock>) {
	$self->{XML} .= $res;
	last if ($res =~ /<\/Result>/);
    }

    if (length($self->{XML}) == 0) {
	$self->{ErrorState} = "HY000";
	$self->{ErrorText} = "Failed read on socket - $!";
	close($self->{_socket}) if ($self->{_Connected});
	$self->{_socket} = undef;
	$self->{_Connected} = 0;
	return undef;
    }
	
    $self->{_executed} = 1;
    $self->{_pos} = 0;

    return 1 if (!$self->{Parse});

    return _parse($self);

    return 1;
}

#
# First
#
sub First {
    my $self = shift;

    _init_state($self);
    if (!$self->{_Connected}) {
	$self->{ErrorState} = "HY000";
	$self->{ErrorText} = "Not connected";
	return undef;
    }
    if (!$self->{_executed}) {
	$self->{ErrorState} = "HY000";
	$self->{ErrorText} = "No SQL executed yet";
	return undef;
    }
    if (!$self->{Parse}) {
	$self->{ErrorState} = "HY000";
	$self->{ErrorText} = "Result not parsed";
	return undef;
    }
    $self->{_pos} = 0;
    return $self->{Rows}->[0];

}
#
# Next
#
sub Next {
    my $self = shift;

    _init_state($self);
    if (!$self->{_Connected}) {
	$self->{ErrorState} = "HY000";
	$self->{ErrorText} = "Not connected";
	return undef;
    }
    if (!$self->{_executed}) {
	$self->{ErrorState} = "HY000";
	$self->{ErrorText} = "No SQL executed yet";
	return undef;
    }
    if (!$self->{Parse}) {
	$self->ErrorState = "HY000";
	$self->ErrorText = "Result not parsed";
	return undef;
    }
    $self->{_pos}++;
    return undef if ($self->{_pos} >= $self->{NumRows});

    return $self->{Rows}->[$self->{_pos}];

}
#
# Last
#
sub Last {
    my $self = shift;

    _init_state($self);
    if (!$self->{_Connected}) {
	$self->{ErrorState} = "HY000";
	$self->{ErrorText} = "Not connected";
	return undef;
    }
    if (!$self->{_executed}) {
	$self->{ErrorState} = "HY000";
	$self->{ErrorText} = "No SQL executed yet";
	return undef;
    }
    if (!$self->{Parse}) {
	$self->ErrorState = "HY000";
	$self->ErrorText = "Result not parsed";
	return undef;
    }
    $self->{_pos} = $self->{NumRows};

    return $self->{Rows}->[$self->{NumRows} - 1];

}

sub _parse {
    require XML::Parser;

    my $self = shift;

    my $parser = new XML::Parser(ErrorContext => 2);
    $parser->setHandlers(Start => \&start_handler,
			 End => \&end_handler,
			 Char  => \&char_handler);
    return 1 if ($parser->parse($self->{XML}));
}

#Formats error message with XML wrapper consistent with other clients
sub FormatErrorString
{
    my $sReturn = "<?xml version=\"1.0\"?>\r\n";
    
    $sReturn .= 
	"<Result State=\"08001\">\r\n" .
	"<Diag Number=\"1\">\r\n" .
	"<State>08001</State>\r\n" .
	"<Native>" . (0 + $!) . "</Native>\r\n" .
	"<Text>" . $_[0] . " - $!</Text>\r\n" .
	"</Diag>\r\n" .
	"</Result>\r\n";

    return($sReturn);
}

#
# Expat start tag handler
#
sub start_handler
{
    my ($p, $elem, %attrs) = @_;
    my $self = $_self;

    if ($elem =~ "Result") {
	if (%attrs) {
	    if (exists($attrs{'State'})) {
		$_self->{ErrorState} = $attrs{'State'}
	    }
	}
    }
    if (($elem =~ "Row") || ($elem =~ "Diagnostic")) {
	my $rh = $self->{_rrhash};

	#print "row now $resultrow\n";
	foreach my $i (keys %$rh) {
	    $rh->{$i}= "";
	}
    }
    if ($elem =~ "Column") {
	$self->{NumColumns}++ if ($self->{NumRows} == 0);
	$self->{_current_column} = $attrs{Name} if exists($attrs{Name});
    } else {
	$self->{_current_column} = $elem;
    }
    my $stk = $self->{_elem_stack};
    push @$stk, $elem;

}  # End of start_handler

#
# Expat end tag handler
#
sub end_handler
{
    my ($p, $elem) = @_;
    my $self = $_self;

    if (($elem =~ "Row") || ($elem =~ "Diagnostic")) {
	my $rh = $self->{_rrhash};

	my %rowhash = %$rh;

	$self->{Rows}[$self->{NumRows}] = \%rowhash;
	$self->{NumRows}++ if ($elem =~ "Row");
    }
    my $stk = $self->{_elem_stack};
    pop @$stk;

}  # End of end_handler

#
# Expat data handler
#
sub char_handler
{
    my ($p, $data) = @_;
    my $self = $_self;

    my $elems = $self->{_elem_stack};
    my $nelems;

    $nelems = @$elems + 0 - 1;

    return if ($elems->[$nelems] =~ "Row");
    return if ($elems->[$nelems] =~ "Result");
    return if ($elems->[$nelems] =~ "Diagnostic");

    my $rh = $self->{_rrhash};

    $rh->{$self->{_current_column}} .= $data;

    if (($self->{ErrorState} ne "00000") && 
	($self->{ErrorText} eq "") &&
	($self->{_current_column} eq "Text")) {
	$self->{ErrorText} = $data;
    }
} # End of char handler

sub _escape {
    my $sql = shift;
    if ($sql =~ /<!\[CDATA\[/) {
	# CDATA - leave it alone.
    } else {
	# format SQL to remove &, <, and >
        $sql =~ s/&/&amp\;/g;
        $sql =~ s/</&lt\;/g;
        $sql =~ s/>/&gt\;/g;
    }
    return $sql;
}

sub _init_state {
    my $s = shift;
    $s->{ErrorState} = "00000";
    $s->{ErrorText} = "";
}
sub _init_execute {
    my $s = shift;

    $s->{NumRows} = 0;
    $s->{NumColumns} = 0;
    $s->{_current_column} = undef;
    $s->{Rows} = ();
    $s->{_elem_stack} = [];
    $s->{_rrhash} = {};
    $s->{_executed} = 0;
    $s->{Request} = "";
    _init_state($s);
}

1;
__END__


=head1 GETTING HELP

Please read the documentation.
Please read the XML-ODBC Server FAQ.

If you need assistance using this module please use one of the
following resources:

If you have a support contract:

Easysoft support
  email - support@easysoft.com
  +44 (0) 1937 863 450

otherwise you can use the newsgroup:

news://news.easysoft.com/easysoft.public.xml-odbc

=head1 DESCRIPTION

This module allows you to issue SQL requests in XML to a local or
remote ODBC driver where the Easysoft XML-ODBC Server is running. The
results are returned in XML although this module can also parse the
XML and provide the results in other formats.

=head1 PROTOCOL

The protocol used by the XML-ODBC Server is a simple request/response
cycle with the requests and responses documented by the Document Type
Definitions (DTD) distributed with the XML-ODBC Server.

=head1 Notation and Conventions

 $h     XMREQUEST handle object
 \%attr reference to a hash of attribute values passed to methods
 $file  filename

=head1 new

You must create a new $h before using any of the other methods.

  $h = ESXMLODBC::XMLREQUEST->new();

=head1 Connect

The Connect method creates a connection between the client 
and server. The Connect method needs two attributes to acheive this

=over 4

=item B<Server>

the name or IP address of the machine where the XML-ODBC Server is
listening.

=item B<Port>

the port when the XML-ODBC Server is listening.

=back

The Port is optional and defaults to 8895.

You may independently set Port and Server before using Connect or
pass a hash reference of connection attributes to Connect.

e.g.

    $h->{Server} = "demo.easysoft.com";
    $h->{Port} = "8895";
    $h->Connect();

or

    $h->Connect({Server=>"demo.easysoft.com", Port=>"8895"});

Connect returns true on successful connection.

=head1  Disconnect

The Disconnect method closes the connection between the client and
the server.

=head1 ExecuteFile (request_filespec)

  $h->ExecuteFile($filespec);

This method allows you to pass an XML request residing in the
specified file directly to the XML-ODBC Server you are connected
to. The format of these XML requests is documented in the DTDs
distributed with the XML-ODBC Server.

A simple example is:

  <?xml version="1.0" encoding="UTF-8" ?>
  <Request>
    <ConnectionString>DSN=demo;UID=demo;PWD=easysoft;</ConnectionString>
    <Sql>select * from oobdist_contact</Sql>
  </Request>

  $h->ExecuteFile("/tmp/request.xml");

=head1 Execute

  $h->Execute();
  $h->Execute($sql);
  $h->Execute($sql, \%attr);

This method allows you to send an SQL request to the XML-ODBC Server.
Two attributes are required:

  ConnectionString - an ODBC connection string defining the ODBC driver
                     and database you want to connect to.

  Sql - this is either passed as the first argument to Execute OR
        set as a reference to array of SQL statements you would like
        executed as part of a transaction (see Transactions).

You can define ConnectionString and Sql before calling Execute by
setting them in the hash reference.

e.g.

    $h->{ConnectionString} = "DSN=demo.easysoft.com";

    $h->{Sql} = ["insert into table values (1)",
                 "update table set columna = 2 where columnb = 1"];

Do not set Sql as above unless you are wanting to use
transactions. Instead pass the SQL are the first argument to Execute:

  $h->Execute("select * from table");

There are various other execution attributes which may be set in
the Execute call by providing a hash reference:

e.g.

  $h->Execute("select * from table", {Parse=>"1"});

See Execution attributes.

=head1 First Next and Last

These methods may be used after using Execute or ExecuteFile.

All these methods return a reference to a hash containing the column
information (column name is the key) or undef.

  $h->First - return the first row in the result set
  $h->Next - return the next row in the result set
  $h->Last - return the last row in the result set

Note that these methods may only be used if the Execution attribute
Parse is set to 1 before the Execute method is used. Note that you
also need Expat and XML::Parser installed (see the READNE.txt).

example:

  $h->Execute("select * from table", {Parse=>"1"});
  $hr = $xoh->First() || die "cannot get first row";
  print join ("," keys %$hr), "\n"; # print column names
  while($hr) {
    my $col;
    foreach $col (keys %$hr) {
      print $hr->{$col}, " ";
    }
    print "\n";
    $hr = $xoh->Next();
  }

=head1 Connection attributes

These attributes may be set directly in the hash off $h or by passing
a hash refernce to the Connect method.

=over 4

=item B<ConnectionString> (string, read/write)

This is the ODBC connection string to be passed to the ODBC API
SQLDriverConnectW. It consists on a number of attribute=value pairs
separated by semicolons. At its simplest a ConnectionString is:

  DSN=system_dsn_name;

where the DSN attribute names the B<SYSTEM> data source name known
to the ODBC driver manager. However, most ODBC drivers define a number of
additional attributes

e.g.

B<UID>

The database username

B<PWD>

The database user's password.

The ODBC driver manager also accepts some specific attibutes like
B<DRIVER> and B<FILEDSN> which can be used to specify the ODBC driver
or a data source defined in a file.

Consult a good ODBC reference and your ODBC driver's documentation
for a list of all the possibilities.

=back

=head1 Execution attributes

These attributes may be set directly in the hash off $h or by passing
a hash reference to the Execute method.

=over 4

=item B<Parse> (boolen, read/write)

If set to 1 causes the resulting XML to be parsed so the First, Next
and Last methods may be used and also sets the NumRows, NumColumns and
Rows (see Execution Results).

=item B<Truncate>

If set to 1 causes the <Truncate> Format element to be sent to the
server which asks it to remove trailing spaces from the end of
SQL_CHAR columns.

=back

=head1 Execution Results

There are a number of execution result attributes which may be
accessed off $h.

=over 4

=item B<XML>

This contains the XML result returned from the XML-ODBC Server.

=item B<NumRows>

This contains the number of rows in the result-set. This is undefined
unless B<Parse> is set to 1 before execution.

=item B<Request>

This is the XML request sent to the XML-ODBC Server. This is only
defined if the Execute method is used.

=item B<NumColumns>

This contains the number of columns in a row in the result-set.  This
is undefined unless B<Parse> is set to 1 before the execution.

=item B<Rows>

This is a reference to an array of hash references. There is one hash
reference for each row in the result set. The keys to the hash
reference are the column names.

    Rows->[0]->{Column1Name}
    Rows->[1]->{Column1Name}
    Rows->[2]->{Column1Name}
    .
    .
    Rows[n]->{ColumnnName}

This is only defined if the Execute method is used.

=back

=head1 Error Handling

In this beta of the XML-ODBC Server the error handling in this module
has not been finalised. Currently, all methods return undef for error
and set B<ErrorState> and B<ErrorText> when an error occurs.

=over 4

=item B<ErrorState>

A 5 character string identifying the error. This is similar to ODBC
states. The string "00000" represents success. Any other error state
string is an error.

=item B<ErrorText>

A description of the error.

=back


If parsing is turned off (see Parse) then the ErrorState will always
be "00000" if the request was sent to the server and a response was
received, even though the XML response suggests the requested action
could not be completed successfully. In this case you need to actually
examine the returned XML to identify the problem.

e.g.

With Parse=0, and executing this request:

  <Request>
    <ConnectionString>DSN=test;</ConnectionString>
    <Sql><![CDATA[select * from tabledoesnotexist]]></Sql>
  </Request>

The returned XML is probably something like this:

  <Result State="S0002">
  <Diagnostic Number="1">
      <State>S0002</State>
      <Native>208</Native>
      <Text>[Microsoft][ODBC SQL Server Driver][SQL Server]
            Invalid object name 'tabledoesnotexist'.
      </Text>
  </Diagnostic>

indicating the table did not exist but the B<Execute> method will
return successfully and B<ErrorState> will be "00000".

To get full B<ErrorState> and B<ErrorText> strings from the returned
XML you need to set Parse to 1. In this case B<ErrorState> is set to
either a client end error state (e.g. socket could not be created) or
to the State attribute in the Result element of the result (this will
always be "00000" for success or the State in the first ODBC
diagnostic). In this way you can tell if your SQL execution was OK
by testing B<ErrorState> eq "00000".


=cut
