n#!/usr/bin/perl -w
#
# Copyright (c) 1993-2002 Easysoft Ltd. All rights reserved.
#
# $Id: xmlodbc_client.pl,v 1.3 2002/07/08 07:15:29 martin Exp $
#
# This is an example perl script using the ESXMLODBC::XMLREQUEST package to
# send a SQL request in XML to the the Easysoft XML-ODBC Server and
# output the result in XML.
#
# Usage:
#
# xmlodbc_client.pl -h server [-p port] 
#                   [-t] [-r] [-V]
#                   [-f file_request] |
#                   [-c connection_string
#                    [-s SQL query] | SQL_statements_in_a_transaction]]
#
# see the usage and command line error text below.
#
# e.g. 1
#
# xmlodbc_client.pl -h myserver.company.co.uk -p 8895 
#                   -c "DSN=test;UID=dbuser;PWD=dbpassword;"
#                   -s "select * from mytable"
#
# which causes the XML-ODBC Server on myserver.company.co.uk listening
# on port 8895 to connect to the test data source with a database username
# and password of dbuser/dbpassword, execute the select statement and returns
# the results in XML.
#
# e.g. 2
#
# xmlodbc_client.pl -h myserver.company.co.uk -p 8895
#                   -c "DSN=test;UID=dbuser;PWD=dbpassword;"
#                   "insert into users values (1,'user@domain')"
#                   "insert into mailing values ('user@domain', 'Yes')"
#
# which causes the XML-ODBC Server on myserver.company.co.uk listening
# on port 8895 to connect to the test data source with a database username
# and password of dbuser/dbpassword, and execute the 2 insert statements.
# If both inserts work they are committed. If either insert fails the
# transaction is rolled back.
#
# e.g. 3
#
# xmlodbc_client.pl -h myserver.company.co.uk -p 8895
#                   -f file.xml
#
# This sends the SQL request in file.xml to the XML-ODBC Server on
# myserver.company.co.uk listening on port 8895 and returns the result
# in XML.
# You can use this method to send configuration requests/instructions
# and statistics requests to the XML-ODBC Server's control port as well
# (see the DTDs and XML example requests for how to do this).
#

use 5.005;
use strict;
use Getopt::Std;

use ESXMLODBC::XMLREQUEST;

sub usage
{
    print <<EOF
xmlodbc_client.pl -h server [-p port] 
                  [-t] [-r] [-V]
                  [-f file_request] |
                  [-c connection_string 
                  [-s sql_query | SQL_statements_in_a_transaction]]

EOF
;
}

my %opts;

if (!getopts('h:p:c:s:f:trV', \%opts)) {
    usage();
    print <<EOF

    -h name of machine where XML-ODBC server is running
    -p the port on server to connect to (usually 8895 or 8896)
    -c ODBC connection string
    -s SQL to run
    -f file containing an XML request
    -t truncate trailing characters in SQL_CHAR columns
       (cannot be used with -f)
    -r display the XML request
       (cannot be used with -f)
    -V print version of XMLODBC
EOF
;
    exit 1;
}

if ($opts{'V'}) {
    print "Version: $ESXMLODBC::VERSION\n";
    exit 0;
}

if (!exists($opts{'h'})) {
    usage();
    die "missing server\n";
}

if (!exists($opts{'c'})) {
    usage();
    die "missing connection";
}

if (!(scalar(@ARGV))) {
    if (!exists($opts{'s'}) && !exists($opts{'f'})) {
	usage();
	die "missing SQL";
    }
}

if (exists($opts{'f'}) && (exists($opts{'t'}) || exists($opts{'r'}))) {
    usage();
    die "-t and -r may not be used with a file request";
}

my $xoh = ESXMLODBC::XMLREQUEST->new();
$xoh->{Port} = $opts{'p'} if ($opts{'p'});

if (!$xoh->Connect({Server => $opts{'h'}})) {
    print "$xoh->{ErrorState} - $xoh->{ErrorText}\n";
    exit 1;
}

$xoh->{Truncate} = 1 if ($opts{'t'});

my $sts;

$xoh->{ConnectionString} = $opts{'c'};

if ($opts{'f'}) {
    $sts = $xoh->ExecuteFile($opts{'f'});
} elsif (scalar(@ARGV)) {
    $xoh->{Sql} = \@ARGV;
    $sts = $xoh->Execute({Parse=>"0"});
} else {
    $sts = $xoh->Execute($opts{'s'}, {Parse=>"0"});
}

if (!$sts) {
    print "$xoh->{ErrorState} - $xoh->{ErrorText}\n";
    exit 1;
}

print $xoh->{Request} if $opts{'r'};

print $xoh->{XML};
